@@ -42,15 +42,7 @@ namespace Mono.Cecil.Signatures {
 		ReflectionReader m_reflectReader;
 		byte [] m_blobData;
 
-		// flyweights
-		IDictionary m_fieldSigs;
-		IDictionary m_propSigs;
-		IDictionary m_typeSpecs;
-		IDictionary m_methodSpecs;
-		IDictionary m_methodsDefSigs;
-		IDictionary m_methodsRefSigs;
-		IDictionary m_localVars;
-		IDictionary m_marshalSpecs;
+		IDictionary m_signatures;
 
 		public SignatureReader (MetadataRoot root, ReflectionReader reflectReader)
 		{
@@ -59,95 +51,84 @@ namespace Mono.Cecil.Signatures {
 
 			m_blobData = m_root.Streams.BlobHeap != null ? m_root.Streams.BlobHeap.Data : new byte [0];
 
-			m_fieldSigs = new Hashtable ();
-			m_propSigs = new Hashtable ();
-			m_methodsDefSigs = new Hashtable ();
-			m_methodsRefSigs = new Hashtable ();
-			m_localVars = new Hashtable ();
-			m_marshalSpecs = new Hashtable ();
+			m_signatures = new Hashtable ();
 		}
 
 		public FieldSig GetFieldSig (uint index)
 		{
-			FieldSig f = m_fieldSigs [index] as FieldSig;
+			FieldSig f = m_signatures [index] as FieldSig;
 			if (f == null) {
 				f = new FieldSig (index);
 				f.Accept (this);
-				m_fieldSigs [index] = f;
+				m_signatures [index] = f;
 			}
 			return f;
 		}
 
 		public PropertySig GetPropSig (uint index)
 		{
-			PropertySig p = m_propSigs [index] as PropertySig;
+			PropertySig p = m_signatures [index] as PropertySig;
 			if (p == null) {
 				p = new PropertySig (index);
 				p.Accept (this);
-				m_propSigs [index] = p;
+				m_signatures [index] = p;
 			}
 			return p;
 		}
 
 		public MethodDefSig GetMethodDefSig (uint index)
 		{
-			MethodDefSig m = m_methodsDefSigs [index] as MethodDefSig;
+			MethodDefSig m = m_signatures [index] as MethodDefSig;
 			if (m == null) {
 				m = new MethodDefSig (index);
 				m.Accept (this);
-				m_methodsDefSigs [index] = m;
+				m_signatures [index] = m;
 			}
 			return m;
 		}
 
 		public MethodRefSig GetMethodRefSig (uint index)
 		{
-			MethodRefSig m = m_methodsRefSigs [index] as MethodRefSig;
+			MethodRefSig m = m_signatures [index] as MethodRefSig;
 			if (m == null) {
 				m = new MethodRefSig (index);
 				m.Accept (this);
-				m_methodsRefSigs [index] = m;
+				m_signatures [index] = m;
 			}
 			return m;
 		}
 
 		public TypeSpec GetTypeSpec (uint index)
 		{
-			TypeSpec ts = null;
-			if (m_typeSpecs == null)
-				m_typeSpecs = new Hashtable ();
-			else
-				ts = m_typeSpecs [index] as TypeSpec;
+			TypeSpec ts = m_signatures [index] as TypeSpec;
 
 			if (ts == null) {
 				ts = ReadTypeSpec (m_blobData, (int) index);
-				m_typeSpecs [index] = ts;
+				m_signatures [index] = ts;
 			}
+
 			return ts;
 		}
 
 		public MethodSpec GetMethodSpec (uint index)
 		{
-			MethodSpec ms = null;
-			if (m_methodSpecs == null)
-				m_methodSpecs = new Hashtable ();
-			else
-				ms = m_methodSpecs [index] as MethodSpec;
+			MethodSpec ms = m_signatures [index] as MethodSpec;
 
 			if (ms == null) {
 				ms = ReadMethodSpec (m_blobData, (int) index);
-				m_methodSpecs [index] = ms;
+				m_signatures [index] = ms;
 			}
+
 			return ms;
 		}
 
 		public LocalVarSig GetLocalVarSig (uint index)
 		{
-			LocalVarSig lv = m_localVars [index] as LocalVarSig;
+			LocalVarSig lv = m_signatures [index] as LocalVarSig;
 			if (lv == null) {
 				lv = new LocalVarSig (index);
 				lv.Accept (this);
-				m_localVars [index] = lv;
+				m_signatures [index] = lv;
 			}
 			return lv;
 		}
@@ -187,11 +168,11 @@ namespace Mono.Cecil.Signatures {
 
 		public MarshalSig GetMarshalSig (uint index)
 		{
-			MarshalSig ms = m_marshalSpecs [index] as MarshalSig;
+			MarshalSig ms = m_signatures [index] as MarshalSig;
 			if (ms == null) {
 				byte [] data = m_root.Streams.BlobHeap.Read (index);
 				ms = ReadMarshalSig (data);
-				m_marshalSpecs [index] = ms;
+				m_signatures [index] = ms;
 			}
 			return ms;
 		}
@@ -314,6 +295,7 @@ namespace Mono.Cecil.Signatures {
 			lv.ByRef = false;
 			int cursor;
 			while (true) {
+				lv.CustomMods = ReadCustomMods (data, start, out start);
 				cursor = start;
 				int current = Utilities.ReadCompressedInteger (data, start, out start);
 				if (current == (int) ElementType.Pinned) // the only possible constraint
@@ -557,8 +539,14 @@ namespace Mono.Cecil.Signatures {
 			int start, length = Utilities.ReadCompressedInteger (m_blobData, pos, out start);
 			byte [] data = new byte [length];
 			Buffer.BlockCopy (m_blobData, start, data, 0, length);
-			return ReadCustomAttrib (new BinaryReader (
-				new MemoryStream (data)), data, ctor);
+			try {
+				return ReadCustomAttrib (new BinaryReader (
+					new MemoryStream (data)), data, ctor);
+			} catch {
+				CustomAttrib ca = new CustomAttrib (ctor);
+				ca.Read = false;
+				return ca;
+			}
 		}
 
 		CustomAttrib ReadCustomAttrib (BinaryReader br, byte [] data, MethodReference ctor)
@@ -581,6 +569,9 @@ namespace Mono.Cecil.Signatures {
 				ca.FixedArgs [i] = ReadFixedArg (data, br, ctor.Parameters [i].ParameterType is ArrayType,
 					ctor.Parameters [i].ParameterType, ref read);
 
+			if (br.BaseStream.Position == br.BaseStream.Length)
+				read = false;
+
 			if (!read) {
 				ca.Read = read;
 				return ca;
@@ -608,6 +599,9 @@ namespace Mono.Cecil.Signatures {
 					return fa;
 				}
 
+				if (param is ArrayType)
+					param = ((ArrayType) param).ElementType;
+
 				fa.Elems = new CustomAttrib.Elem [fa.NumElem];
 				for (int i = 0; i < fa.NumElem; i++)
 					fa.Elems [i] = ReadElem (data, br, param, ref read);
@@ -646,7 +640,7 @@ namespace Mono.Cecil.Signatures {
 
 			length = Utilities.ReadCompressedInteger (data, (int) br.BaseStream.Position, out next);
 			br.BaseStream.Position = next;
-			
+
 			// COMPACT FRAMEWORK NOTE: Encoding.GetString(byte[]) is not supported.
 			byte [] bytes = br.ReadBytes (length);
 			na.FieldOrPropName = Encoding.UTF8.GetString (bytes, 0, bytes.Length);
