@@ -62,7 +62,7 @@ namespace Mono.Cecil {
 		protected GenericInstanceMethod [] m_methodSpecs;
 
 		bool m_isCorlib;
-		IAssemblyNameReference m_corlib;
+		AssemblyNameReference m_corlib;
 
 		protected SignatureReader m_sigReader;
 		protected CodeReader m_codeReader;
@@ -75,6 +75,10 @@ namespace Mono.Cecil {
 			get { return m_sigReader; }
 		}
 
+		public MetadataTableReader TableReader {
+			get { return m_tableReader; }
+		}
+
 		public CodeReader Code {
 			get { return m_codeReader; }
 		}
@@ -151,8 +155,14 @@ namespace Mono.Cecil {
 				TypeReference declaringType = GetTypeDefOrRef (mrefRow.Class, context);
 				GenericContext nc = context.Clone ();
 
-				if (mrefRow.Class.TokenType == TokenType.TypeSpec)
-					nc.Type = declaringType;
+				if (declaringType is GenericInstanceType) {
+					TypeReference ct = declaringType;
+					while (ct is GenericInstanceType)
+						ct = (ct as GenericInstanceType).ElementType;
+
+					nc.Type = ct;
+					nc.AllowCreation = ct.GetType () == typeof (TypeReference);
+				}
 
 				if (sig is FieldSig) {
 					FieldSig fs = sig as FieldSig;
@@ -278,15 +288,16 @@ namespace Mono.Cecil {
 		public TypeReference SearchCoreType (string fullName)
 		{
 			if (m_isCorlib)
-				return m_module.Types [fullName] as TypeReference;
+				return m_module.Types [fullName];
 
-			TypeReference coreType =  m_module.TypeReferences [fullName] as TypeReference;
+			TypeReference coreType =  m_module.TypeReferences [fullName];
 			if (coreType == null) {
 				if (m_corlib == null) {
-					foreach (IAssemblyNameReference ar in m_module.AssemblyReferences) {
-						if (ar.Name == Constants.Corlib)
+					foreach (AssemblyNameReference ar in m_module.AssemblyReferences) {
+						if (ar.Name == Constants.Corlib) {
 							m_corlib = ar;
-						break;
+							break;
+						}
 					}
 				}
 
@@ -355,7 +366,7 @@ namespace Mono.Cecil {
 		public override void VisitTypeDefinitionCollection (TypeDefinitionCollection types)
 		{
 			// type def reading
-			TypeDefTable typesTable = m_tHeap [typeof (TypeDefTable)] as TypeDefTable;
+			TypeDefTable typesTable = m_tableReader.GetTypeDefTable ();
 			m_typeDefs = new TypeDefinition [typesTable.Rows.Count];
 			for (int i = 0; i < typesTable.Rows.Count; i++) {
 				TypeDefRow type = typesTable [i];
@@ -369,8 +380,8 @@ namespace Mono.Cecil {
 			}
 
 			// nested types
-			NestedClassTable nested = m_tHeap [typeof (NestedClassTable)] as NestedClassTable;
-			if (m_tHeap.HasTable (typeof(NestedClassTable))) {
+			if (m_tHeap.HasTable (NestedClassTable.RId)) {
+				NestedClassTable nested = m_tableReader.GetNestedClassTable ();
 				for (int i = 0; i < nested.Rows.Count; i++) {
 					NestedClassRow row = nested [i];
 
@@ -385,44 +396,13 @@ namespace Mono.Cecil {
 				types.Add (type);
 
 			// type ref reading
-			if (m_tHeap.HasTable(typeof (TypeRefTable))) {
-				TypeRefTable typesRef = m_tHeap [typeof (TypeRefTable)] as TypeRefTable;
+			if (m_tHeap.HasTable (TypeRefTable.RId)) {
+				TypeRefTable typesRef = m_tableReader.GetTypeRefTable ();
 
 				m_typeRefs = new TypeReference [typesRef.Rows.Count];
 
-				for (int i = 0; i < typesRef.Rows.Count; i++) {
-					TypeRefRow type = typesRef [i];
-					IMetadataScope scope = null;
-					TypeReference parent = null;
-					switch (type.ResolutionScope.TokenType) {
-					case TokenType.AssemblyRef :
-						scope = m_module.AssemblyReferences [(int) type.ResolutionScope.RID - 1];
-						break;
-					case TokenType.ModuleRef :
-						scope = m_module.ModuleReferences [(int) type.ResolutionScope.RID - 1];
-						break;
-					case TokenType.Module :
-						scope = m_module.Assembly.Modules [(int) type.ResolutionScope.RID - 1];
-						break;
-					case TokenType.TypeRef :
-						parent = GetTypeRefAt (type.ResolutionScope.RID);
-						scope = parent.Scope;
-						break;
-					}
-
-					TypeReference t = new TypeReference (
-						m_root.Streams.StringsHeap [type.Name],
-						m_root.Streams.StringsHeap [type.Namespace],
-						scope);
-					t.MetadataToken = MetadataToken.FromMetadataRow (TokenType.TypeRef, i);
-
-					if (parent != null)
-						t.DeclaringType = parent;
-
-					m_typeRefs [i] = t;
-					m_module.TypeReferences.Add (t);
-				}
-
+				for (int i = 0; i < typesRef.Rows.Count; i++)
+					AddTypeRef (typesRef, i);
 			} else
 				m_typeRefs = new TypeReference [0];
 
@@ -444,9 +424,51 @@ namespace Mono.Cecil {
 			ReadMemberReferences ();
 		}
 
+		void AddTypeRef (TypeRefTable typesRef, int i)
+		{
+			// Check if index has been already added.
+			if (m_typeRefs [i] != null)
+				return;
+
+			TypeRefRow type = typesRef [i];
+			IMetadataScope scope = null;
+			TypeReference parent = null;
+
+			if (type.ResolutionScope != MetadataToken.Zero) {
+				switch (type.ResolutionScope.TokenType) {
+				case TokenType.AssemblyRef:
+					scope = m_module.AssemblyReferences [(int) type.ResolutionScope.RID - 1];
+					break;
+				case TokenType.ModuleRef:
+					scope = m_module.ModuleReferences [(int) type.ResolutionScope.RID - 1];
+					break;
+				case TokenType.Module:
+					scope = m_module.Assembly.Modules [(int) type.ResolutionScope.RID - 1];
+					break;
+				case TokenType.TypeRef:
+					AddTypeRef (typesRef, (int) type.ResolutionScope.RID - 1);
+					parent = GetTypeRefAt (type.ResolutionScope.RID);
+					scope = parent.Scope;
+					break;
+				}
+			}
+
+			TypeReference t = new TypeReference (
+				m_root.Streams.StringsHeap [type.Name],
+				m_root.Streams.StringsHeap [type.Namespace],
+				scope);
+			t.MetadataToken = MetadataToken.FromMetadataRow (TokenType.TypeRef, i);
+
+			if (parent != null)
+				t.DeclaringType = parent;
+
+			m_typeRefs [i] = t;
+			m_module.TypeReferences.Add (t);
+		}
+
 		void ReadTypeSpecs ()
 		{
-			if (!m_tHeap.HasTable (typeof (TypeSpecTable)))
+			if (!m_tHeap.HasTable (TypeSpecTable.RId))
 				return;
 
 			TypeSpecTable tsTable = m_tableReader.GetTypeSpecTable ();
@@ -455,7 +477,7 @@ namespace Mono.Cecil {
 
 		void ReadMethodSpecs ()
 		{
-			if (!m_tHeap.HasTable (typeof (MethodSpecTable)))
+			if (!m_tHeap.HasTable (MethodSpecTable.RId))
 				return;
 
 			MethodSpecTable msTable = m_tableReader.GetMethodSpecTable ();
@@ -464,7 +486,7 @@ namespace Mono.Cecil {
 
 		void ReadGenericParameters ()
 		{
-			if (!m_tHeap.HasTable (typeof (GenericParamTable)))
+			if (!m_tHeap.HasTable (GenericParamTable.RId))
 				return;
 
 			GenericParamTable gpTable = m_tableReader.GetGenericParamTable ();
@@ -479,9 +501,10 @@ namespace Mono.Cecil {
 				else
 					throw new ReflectionException ("Unknown owner type for generic parameter");
 
-				GenericParameter gp = new GenericParameter ((int) gpRow.Number, owner);
+				GenericParameter gp = new GenericParameter (gpRow.Number, owner);
 				gp.Attributes = gpRow.Flags;
 				gp.Name = MetadataRoot.Streams.StringsHeap [gpRow.Name];
+				gp.MetadataToken = MetadataToken.FromMetadataRow (TokenType.GenericParam, i);
 
 				owner.GenericParameters.Add (gp);
 				m_genericParameters [i] = gp;
@@ -490,13 +513,14 @@ namespace Mono.Cecil {
 
 		void ReadAllFields ()
 		{
-			TypeDefTable tdefTable = m_tHeap [typeof (TypeDefTable)] as TypeDefTable;
-			if (!m_tHeap.HasTable(typeof (FieldTable))) {
+			TypeDefTable tdefTable = m_tableReader.GetTypeDefTable ();
+
+			if (!m_tHeap.HasTable(FieldTable.RId)) {
 				m_fields = new FieldDefinition [0];
 				return;
 			}
 
-			FieldTable fldTable = m_tHeap [typeof (FieldTable)] as FieldTable;
+			FieldTable fldTable = m_tableReader.GetFieldTable ();
 			m_fields = new FieldDefinition [fldTable.Rows.Count];
 
 			for (int i = 0; i < m_typeDefs.Length; i++) {
@@ -529,7 +553,7 @@ namespace Mono.Cecil {
 
 		void ReadMethods ()
 		{
-			if (!m_tHeap.HasTable (typeof (MethodTable))) {
+			if (!m_tHeap.HasTable (MethodTable.RId)) {
 				m_meths = new MethodDefinition [0];
 				return;
 			}
@@ -554,14 +578,14 @@ namespace Mono.Cecil {
 		{
 			TypeDefTable tdefTable = m_tableReader.GetTypeDefTable ();
 
-			if (!m_tHeap.HasTable (typeof (MethodTable))) {
+			if (!m_tHeap.HasTable (MethodTable.RId)) {
 				m_meths = new MethodDefinition [0];
 				return;
 			}
 
 			MethodTable methTable = m_tableReader.GetMethodTable ();
 			ParamTable paramTable = m_tableReader.GetParamTable ();
-			if (!m_tHeap.HasTable (typeof (ParamTable)))
+			if (!m_tHeap.HasTable (ParamTable.RId))
 				m_parameters = new ParameterDefinition [0];
 			else
 				m_parameters = new ParameterDefinition [paramTable.Rows.Count];
@@ -606,7 +630,11 @@ namespace Mono.Cecil {
 						pRow = paramTable [start];
 
 					if (pRow != null && pRow.Sequence == 0) { // ret type
-						retparam = new ParameterDefinition (string.Empty, 0, (ParamAttributes) 0, null);
+						retparam = new ParameterDefinition (
+							m_root.Streams.StringsHeap [pRow.Name],
+							0,
+							pRow.Flags,
+							null);
 						retparam.Method = mdef;
 						m_parameters [start] = retparam;
 						start++;
@@ -638,7 +666,7 @@ namespace Mono.Cecil {
 					}
 
 					mdef.ReturnType = GetMethodReturnType (msig, context);
-					MethodReturnType mrt = mdef.ReturnType as MethodReturnType;
+					MethodReturnType mrt = mdef.ReturnType;
 					mrt.Method = mdef;
 					if (retparam != null) {
 						mrt.Parameter = retparam;
@@ -654,7 +682,7 @@ namespace Mono.Cecil {
 
 		void ReadMemberReferences ()
 		{
-			if (!m_tHeap.HasTable (typeof (MemberRefTable)))
+			if (!m_tHeap.HasTable (MemberRefTable.RId))
 				return;
 
 			MemberRefTable mrefTable = m_tableReader.GetMemberRefTable ();
@@ -663,12 +691,12 @@ namespace Mono.Cecil {
 
 		public override void VisitExternTypeCollection (ExternTypeCollection externs)
 		{
-			ExternTypeCollection ext = externs as ExternTypeCollection;
+			ExternTypeCollection ext = externs;
 
-			if (!m_tHeap.HasTable (typeof (ExportedTypeTable)))
+			if (!m_tHeap.HasTable (ExportedTypeTable.RId))
 				return;
 
-			ExportedTypeTable etTable = m_tHeap [typeof (ExportedTypeTable)] as ExportedTypeTable;
+			ExportedTypeTable etTable = m_tableReader.GetExportedTypeTable ();
 			TypeReference [] buffer = new TypeReference [etTable.Rows.Count];
 
 			for (int i = 0; i < etTable.Rows.Count; i++) {
@@ -689,7 +717,7 @@ namespace Mono.Cecil {
 				if (etRow.Implementation.TokenType != TokenType.ExportedType)
 					continue;
 
-				ITypeReference owner = buffer [etRow.Implementation.RID - 1];
+				TypeReference owner = buffer [etRow.Implementation.RID - 1];
 				string name = m_root.Streams.StringsHeap [etRow.TypeName];
 				buffer [i] = m_module.TypeReferences [string.Concat (owner.FullName, '/', name)];
 			}
@@ -749,7 +777,7 @@ namespace Mono.Cecil {
 			ParamAttributes attrs, Param psig, GenericContext context)
 		{
 			ParameterDefinition ret = new ParameterDefinition (name, sequence, attrs, null);
-			TypeReference paramType = null;
+			TypeReference paramType;
 
 			if (psig.ByRef)
 				paramType = new ReferenceType (GetTypeRefFromSig (psig.Type, context));
@@ -793,7 +821,7 @@ namespace Mono.Cecil {
 				MarshalSig.CustomMarshaler cmsig = (MarshalSig.CustomMarshaler) ms.Spec;
 				cmd.Guid = cmsig.Guid.Length > 0 ? new Guid (cmsig.Guid) : new Guid ();
 				cmd.UnmanagedType = cmsig.UnmanagedType;
-				cmd.ManagedType = m_module.Types [cmsig.ManagedType];
+				cmd.ManagedType = cmsig.ManagedType;
 				cmd.Cookie = cmsig.Cookie;
 				return cmd;
 			} else if (ms.Spec is MarshalSig.FixedArray) {
@@ -815,12 +843,12 @@ namespace Mono.Cecil {
 			}
 		}
 
-		protected TypeReference GetModifierType (CustomMod [] cmods, TypeReference type)
+		public TypeReference GetModifierType (CustomMod [] cmods, TypeReference type)
 		{
 			TypeReference ret = type;
 			for (int i = cmods.Length - 1; i >= 0; i--) {
 				CustomMod cmod = cmods [i];
-				TypeReference modType = null;
+				TypeReference modType;
 
 				if (cmod.TypeDefOrRef.TokenType == TokenType.TypeDef)
 					modType = GetTypeDefAt (cmod.TypeDefOrRef.RID);
@@ -837,7 +865,7 @@ namespace Mono.Cecil {
 
 		public MethodReturnType GetMethodReturnType (MethodSig msig, GenericContext context)
 		{
-			TypeReference retType = null;
+			TypeReference retType;
 			if (msig.RetType.Void)
 				retType = SearchCoreType (Constants.Void);
 			else if (msig.RetType.ByRef)
@@ -927,6 +955,9 @@ namespace Mono.Cecil {
 				return fnptr;
 			case ElementType.Var:
 				VAR var = t as VAR;
+				if (context.AllowCreation)
+					CheckGenericParameters (context, var);
+
 				if (context.Type is GenericInstanceType)
 					return (context.Type as GenericInstanceType).GenericArguments [var.Index];
 				else
@@ -942,20 +973,9 @@ namespace Mono.Cecil {
 				GenericInstanceType instance = new GenericInstanceType (GetTypeDefOrRef (ginst.Type, context));
 				instance.IsValueType = ginst.ValueType;
 
-				GenericContext ctx = context.Clone ();
-
-				for (int i = 0; i < ginst.Signature.Arity; i++) {
-					if (ginst.Signature.Types [i] is VAR && instance.ElementType.GetType () == typeof (TypeReference)) {
-						VAR v = ginst.Signature.Types [i] as VAR;
-						if (v.Index >= instance.ElementType.GenericParameters.Count)
-							for (int j = 0; j <= v.Index; j++)
-								instance.ElementType.GenericParameters.Add (new GenericParameter (j, instance.ElementType));
-						ctx.Type = instance.ElementType;
-					}
-				}
-
 				for (int i = 0; i < ginst.Signature.Arity; i++)
-					instance.GenericArguments.Add (GetTypeRefFromSig (ginst.Signature.Types [i], ctx));
+					instance.GenericArguments.Add (GetTypeRefFromSig (
+						ginst.Signature.Types [i], context));
 
 				return instance;
 			default:
@@ -964,6 +984,12 @@ namespace Mono.Cecil {
 			return null;
 		}
 
+		void CheckGenericParameters (GenericContext context, VAR v)
+		{
+			for (int i = context.Type.GenericParameters.Count; i <= v.Index; i++)
+				context.Type.GenericParameters.Add (new GenericParameter (i, context.Type));
+		}
+
 		protected object GetConstant (uint pos, ElementType elemType)
 		{
 			byte [] constant = m_root.Streams.BlobHeap.Read (pos);
@@ -1007,10 +1033,37 @@ namespace Mono.Cecil {
 
 		protected void SetInitialValue (FieldDefinition field)
 		{
-			if (field.RVA != RVA.Zero && field.FieldType is ITypeDefinition) {
-				BinaryReader br = this.Module.ImageReader.MetadataReader.GetDataReader (field.RVA);
-				field.InitialValue = br.ReadBytes (
-					(int) (field.FieldType as ITypeDefinition).LayoutInfo.ClassSize);
+			int size = 0;
+			switch (field.FieldType.FullName) {
+			case Constants.Byte:
+			case Constants.SByte:
+				size = 1;
+				break;
+			case Constants.Int16:
+			case Constants.UInt16:
+			case Constants.Char:
+				size = 2;
+				break;
+			case Constants.Int32:
+			case Constants.UInt32:
+			case Constants.Single:
+				size = 4;
+				break;
+			case Constants.Int64:
+			case Constants.UInt64:
+			case Constants.Double:
+				size = 8;
+				break;
+			default:
+				TypeDefinition fieldType = field.FieldType as TypeDefinition;
+				if (fieldType != null)
+					size = (int) fieldType.ClassSize;
+				break;
+			}
+
+			if (size > 0 && field.RVA != RVA.Zero) {
+				BinaryReader br = m_reader.MetadataReader.GetDataReader (field.RVA);
+				field.InitialValue = br == null ? new byte [size] : br.ReadBytes (size);
 			} else
 				field.InitialValue = new byte [0];
 		}
