@@ -30,6 +30,7 @@ namespace Mono.Cecil {
 
 	using System;
 	using System.Collections;
+	using System.Globalization;
 	using System.IO;
 	using System.Text;
 
@@ -48,8 +49,11 @@ namespace Mono.Cecil {
 		MetadataTableWriter m_tableWriter;
 		MetadataRowWriter m_rowWriter;
 
+		IList m_typeDefStack;
 		IList m_methodStack;
 		IList m_fieldStack;
+		IList m_genericParamStack;
+		IDictionary m_typeSpecCache;
 
 		uint m_methodIndex;
 		uint m_fieldIndex;
@@ -100,8 +104,11 @@ namespace Mono.Cecil {
 		{
 			m_mod = mod;
 
+			m_typeDefStack = new ArrayList ();
 			m_methodStack = new ArrayList ();
 			m_fieldStack = new ArrayList ();
+			m_genericParamStack = new ArrayList ();
+			m_typeSpecCache = new Hashtable ();
 
 			m_methodIndex = 1;
 			m_fieldIndex = 1;
@@ -112,7 +119,7 @@ namespace Mono.Cecil {
 			m_constWriter = new MemoryBinaryWriter ();
 		}
 
-		public ITypeReference GetCoreType (string name)
+		public TypeReference GetCoreType (string name)
 		{
 			return m_mod.Controller.Reader.SearchCoreType (name);
 		}
@@ -137,20 +144,23 @@ namespace Mono.Cecil {
 			return (uint) m_mod.ModuleReferences.IndexOf (modRef) + 1;
 		}
 
-		bool IsTypeSpec (ITypeReference type)
+		bool IsTypeSpec (TypeReference type)
 		{
-			return type is IArrayType || type is IFunctionPointerType ||
-				type is IPointerType || type is IGenericInstance;
+			return type is TypeSpecification || type is GenericParameter;
 		}
 
-		public MetadataToken GetTypeDefOrRefToken (ITypeReference type)
+		public MetadataToken GetTypeDefOrRefToken (TypeReference type)
 		{
 			if (IsTypeSpec (type)) {
+				uint sig = m_sigWriter.AddTypeSpec (GetTypeSpecSig (type));
+				if (m_typeSpecCache.Contains (sig))
+					return (m_typeSpecCache [sig] as TypeReference).MetadataToken;
+
 				TypeSpecTable tsTable = m_tableWriter.GetTypeSpecTable ();
-				TypeSpecRow tsRow = m_rowWriter.CreateTypeSpecRow (
-					m_sigWriter.AddTypeSpec (GetTypeSpecSig (type)));
+				TypeSpecRow tsRow = m_rowWriter.CreateTypeSpecRow (sig);
 				tsTable.Rows.Add (tsRow);
 				type.MetadataToken = new MetadataToken (TokenType.TypeSpec, (uint) tsTable.Rows.Count);
+				m_typeSpecCache [sig] = type;
 				return type.MetadataToken;
 			} else if (type != null)
 				return type.MetadataToken;
@@ -160,10 +170,18 @@ namespace Mono.Cecil {
 
 		public MetadataToken GetMethodSpecToken (GenericInstanceMethod gim)
 		{
+			uint sig = m_sigWriter.AddMethodSpec (GetMethodSpecSig (gim));
 			MethodSpecTable msTable = m_tableWriter.GetMethodSpecTable ();
+
+			for (int i = 0; i < msTable.Rows.Count; i++) {
+				MethodSpecRow row = msTable [i];
+				if (row.Method == gim.ElementMethod.MetadataToken && row.Instantiation == sig)
+					return MetadataToken.FromMetadataRow (TokenType.MethodSpec, i);
+			}
+
 			MethodSpecRow msRow = m_rowWriter.CreateMethodSpecRow (
 				gim.ElementMethod.MetadataToken,
-				m_sigWriter.AddMethodSpec (GetMethodSpecSig (gim)));
+				sig);
 			msTable.Rows.Add (msRow);
 			gim.MetadataToken = new MetadataToken (TokenType.MethodSpec, (uint) msTable.Rows.Count);
 			return gim.MetadataToken;
@@ -171,38 +189,31 @@ namespace Mono.Cecil {
 
 		public override void VisitModuleDefinition (ModuleDefinition mod)
 		{
-			// ensure that everything is loaded before writing
-			foreach (TypeDefinition type in mod.Types) {
-				foreach (MethodDefinition meth in type.Methods)
-					meth.LoadBody ();
-				foreach (MethodDefinition ctor in type.Constructors)
-					ctor.LoadBody ();
-			}
+			mod.FullLoad ();
 		}
 
 		public override void VisitTypeDefinitionCollection (TypeDefinitionCollection types)
 		{
-			ArrayList orderedTypes = new ArrayList (types.Count);
 			TypeDefTable tdTable = m_tableWriter.GetTypeDefTable ();
 
 			if (types [Constants.ModuleType] == null)
 				types.Add (new TypeDefinition (
-						Constants.ModuleType, string.Empty, (TypeAttributes) 0));
+						Constants.ModuleType, string.Empty, TypeAttributes.NotPublic));
 
-			foreach (ITypeDefinition t in types)
-				orderedTypes.Add (t);
+			foreach (TypeDefinition t in types)
+				m_typeDefStack.Add (t);
 
-			orderedTypes.Sort (TableComparers.TypeDef.Instance);
+			(m_typeDefStack as ArrayList).Sort (TableComparers.TypeDef.Instance);
 
-			for (int i = 0; i < orderedTypes.Count; i++) {
-				TypeDefinition t = (TypeDefinition) orderedTypes [i];
+			for (int i = 0; i < m_typeDefStack.Count; i++) {
+				TypeDefinition t = (TypeDefinition) m_typeDefStack [i];
 				if (t.Module.Assembly != m_mod.Assembly)
 					throw new ReflectionException ("A type as not been correctly imported");
 
 				t.MetadataToken = new MetadataToken (TokenType.TypeDef, (uint) (i + 1));
 			}
 
-			foreach (TypeDefinition t in orderedTypes) {
+			foreach (TypeDefinition t in m_typeDefStack) {
 				TypeDefRow tdRow = m_rowWriter.CreateTypeDefRow (
 					t.Attributes,
 					m_mdWriter.AddString (t.Name),
@@ -212,14 +223,16 @@ namespace Mono.Cecil {
 					0);
 
 				tdTable.Rows.Add (tdRow);
-
-				if (t.LayoutInfo.HasLayoutInfo)
-					WriteLayout (t);
 			}
+		}
 
-			for (int i = 0; i < orderedTypes.Count; i++) {
+		public void CompleteTypeDefinitions ()
+		{
+			TypeDefTable tdTable = m_tableWriter.GetTypeDefTable ();
+
+			for (int i = 0; i < m_typeDefStack.Count; i++) {
 				TypeDefRow tdRow = tdTable [i];
-				TypeDefinition t = (TypeDefinition) orderedTypes [i];
+				TypeDefinition t = (TypeDefinition) m_typeDefStack [i];
 				tdRow.FieldList = m_fieldIndex;
 				tdRow.MethodList = m_methodIndex;
 				foreach (FieldDefinition field in t.Fields)
@@ -228,20 +241,32 @@ namespace Mono.Cecil {
 					VisitMethodDefinition (ctor);
 				foreach (MethodDefinition meth in t.Methods)
 					VisitMethodDefinition (meth);
+
+				if (t.HasLayoutInfo)
+					WriteLayout (t);
 			}
 
-			foreach (FieldDefinition field in m_fieldStack)
+			foreach (FieldDefinition field in m_fieldStack) {
 				VisitCustomAttributeCollection (field.CustomAttributes);
+				if (field.MarshalSpec != null)
+					VisitMarshalSpec (field.MarshalSpec);
+			}
 
 			foreach (MethodDefinition meth in m_methodStack) {
+				VisitCustomAttributeCollection (meth.ReturnType.CustomAttributes);
+				foreach (ParameterDefinition param in meth.Parameters)
+					VisitCustomAttributeCollection (param.CustomAttributes);
+				VisitGenericParameterCollection (meth.GenericParameters);
 				VisitOverrideCollection (meth.Overrides);
 				VisitCustomAttributeCollection (meth.CustomAttributes);
 				VisitSecurityDeclarationCollection (meth.SecurityDeclarations);
-				if (meth.PInvokeInfo != null)
+				if (meth.PInvokeInfo != null) {
+					meth.Attributes |= MethodAttributes.PInvokeImpl;
 					VisitPInvokeInfo (meth.PInvokeInfo);
+				}
 			}
 
-			foreach (TypeDefinition t in orderedTypes)
+			foreach (TypeDefinition t in m_typeDefStack)
 				t.Accept (this);
 		}
 
@@ -295,10 +320,10 @@ namespace Mono.Cecil {
 			MemberRefTable mrTable = m_tableWriter.GetMemberRefTable ();
 			foreach (MemberReference member in members) {
 				uint sig = 0;
-				if (member is IFieldReference)
-					sig = m_sigWriter.AddFieldSig (GetFieldSig (member as IFieldReference));
-				else if (member is IMethodReference)
-					sig = m_sigWriter.AddMethodRefSig (GetMethodRefSig (member as IMethodReference));
+				if (member is FieldReference)
+					sig = m_sigWriter.AddFieldSig (GetFieldSig (member as FieldReference));
+				else if (member is MethodReference)
+					sig = m_sigWriter.AddMethodRefSig (GetMethodRefSig (member as MethodReference));
 				MemberRefRow mrRow = m_rowWriter.CreateMemberRefRow (
 					GetTypeDefOrRefToken (member.DeclaringType),
 					m_mdWriter.AddString (member.Name),
@@ -315,31 +340,8 @@ namespace Mono.Cecil {
 			if (parameters.Count == 0)
 				return;
 
-			GenericParamTable gpTable = m_tableWriter.GetGenericParamTable ();
-			GenericParamConstraintTable gpcTable = m_tableWriter.GetGenericParamConstraintTable ();
-
-			for (int i = 0; i < parameters.Count; i++) {
-
-				GenericParameter gp = parameters [i];
-				GenericParamRow gpRow = m_rowWriter.CreateGenericParamRow (
-					(ushort) i,
-					gp.Attributes,
-					gp.Owner.MetadataToken,
-					m_mdWriter.AddString (gp.Name));
-
-				gpTable.Rows.Add (gpRow);
-
-				if (gp.Constraints.Count == 0)
-					continue;
-
-				foreach (TypeReference constraint in gp.Constraints) {
-					GenericParamConstraintRow gpcRow = m_rowWriter.CreateGenericParamConstraintRow (
-						(uint) gpTable.Rows.Count,
-						GetTypeDefOrRefToken (constraint));
-
-					gpcTable.Rows.Add (gpcRow);
-				}
-			}
+			foreach (GenericParameter gp in parameters)
+				m_genericParamStack.Add (gp);
 		}
 
 		public override void VisitInterfaceCollection (InterfaceCollection interfaces)
@@ -375,7 +377,7 @@ namespace Mono.Cecil {
 			MethodImplTable miTable = m_tableWriter.GetMethodImplTable ();
 			foreach (MethodReference ov in meths) {
 				MethodImplRow miRow = m_rowWriter.CreateMethodImplRow (
-					GetRidFor (meths.Container.DeclaringType as ITypeDefinition),
+					GetRidFor (meths.Container.DeclaringType as TypeDefinition),
 					new MetadataToken (TokenType.Method, GetRidFor (meths.Container)),
 					ov.MetadataToken);
 
@@ -405,16 +407,33 @@ namespace Mono.Cecil {
 
 			ushort seq = 1;
 			ParamTable pTable = m_tableWriter.GetParamTable ();
-			foreach (ParameterDefinition param in parameters) {
-				ParamRow pRow = m_rowWriter.CreateParamRow (
-					param.Attributes,
-					seq++,
-					m_mdWriter.AddString (param.Name));
+			foreach (ParameterDefinition param in parameters)
+				InsertParameter (pTable, param, seq++);
+		}
 
-				pTable.Rows.Add (pRow);
-				param.MetadataToken = new MetadataToken (TokenType.Param, (uint) pTable.Rows.Count);
-				m_paramIndex++;
-			}
+		void InsertParameter (ParamTable pTable, ParameterDefinition param, ushort seq)
+		{
+			ParamRow pRow = m_rowWriter.CreateParamRow (
+				param.Attributes,
+				seq,
+				m_mdWriter.AddString (param.Name));
+
+			pTable.Rows.Add (pRow);
+			param.MetadataToken = new MetadataToken (TokenType.Param, (uint) pTable.Rows.Count);
+
+			if (param.MarshalSpec != null)
+				param.MarshalSpec.Accept (this);
+
+			if (param.HasConstant)
+				WriteConstant (param, param.ParameterType);
+
+			m_paramIndex++;
+		}
+
+		bool RequiresParameterRow (MethodReturnType mrt)
+		{
+			return mrt.HasConstant || mrt.MarshalSpec != null ||
+				mrt.CustomAttributes.Count > 0 || mrt.Parameter.Attributes != (ParamAttributes) 0;
 		}
 
 		public override void VisitMethodDefinition (MethodDefinition method)
@@ -433,17 +452,8 @@ namespace Mono.Cecil {
 			method.MetadataToken = new MetadataToken (TokenType.Method, (uint) mTable.Rows.Count);
 			m_methodIndex++;
 
-			if (method.ReturnType.CustomAttributes.Count > 0 || method.ReturnType.MarshalSpec != null) {
-				ParameterDefinition param = (method.ReturnType as MethodReturnType).Parameter;
-				ParamTable pTable = m_tableWriter.GetParamTable ();
-				ParamRow pRow = m_rowWriter.CreateParamRow (
-					param.Attributes,
-					0,
-					0);
-
-				pTable.Rows.Add (pRow);
-				param.MetadataToken = new MetadataToken (TokenType.Param, (uint) pTable.Rows.Count);
-			}
+			if (RequiresParameterRow (method.ReturnType))
+				InsertParameter (m_tableWriter.GetParamTable (), method.ReturnType.Parameter, 0);
 
 			VisitParameterDefinitionCollection (method.Parameters);
 		}
@@ -510,9 +520,9 @@ namespace Mono.Cecil {
 			m_fieldIndex++;
 
 			if (field.HasConstant)
-				WriteConstant (field, field.MetadataToken, field.FieldType);
+				WriteConstant (field, field.FieldType);
 
-			if (field.LayoutInfo.HasLayoutInfo)
+			if (field.HasLayoutInfo)
 				WriteLayout (field);
 
 			m_fieldStack.Add (field);
@@ -549,34 +559,24 @@ namespace Mono.Cecil {
 			if (property.SetMethod != null)
 				WriteSemantic (MethodSemanticsAttributes.Setter, property, property.SetMethod);
 
+			if (property.HasConstant)
+				WriteConstant (property, property.PropertyType);
+
 			m_propertyIndex++;
 		}
 
-		private MetadataToken GetMetadataToken (IHasSecurity container)
-		{
-			if (container is IAssemblyDefinition)
-				return new MetadataToken (TokenType.Assembly, 1);
-
-			IMetadataTokenProvider provider = (container as IMetadataTokenProvider);
-			if (container == null)
-				throw new ReflectionException ("Unknown Security Declaration parent");
-
-			return provider.MetadataToken;
-		}
-
 		public override void VisitSecurityDeclarationCollection (SecurityDeclarationCollection secDecls)
 		{
 			if (secDecls.Count == 0)
 				return;
 
-			MetadataToken parent = GetMetadataToken (secDecls.Container);
 			DeclSecurityTable dsTable = m_tableWriter.GetDeclSecurityTable ();
 			foreach (SecurityDeclaration secDec in secDecls) {
 				DeclSecurityRow dsRow = m_rowWriter.CreateDeclSecurityRow (
 					secDec.Action,
-					parent,
-					m_mdWriter.AddBlob (
-						m_mod.GetAsByteArray (secDec)));
+					secDecls.Container.MetadataToken,
+					m_mdWriter.AddBlob (secDec.IsReadable ?
+						m_mod.GetAsByteArray (secDec) : secDec.Blob));
 
 				dsTable.Rows.Add (dsRow);
 			}
@@ -590,9 +590,9 @@ namespace Mono.Cecil {
 			CustomAttributeTable caTable = m_tableWriter.GetCustomAttributeTable ();
 			foreach (CustomAttribute ca in customAttrs) {
 				MetadataToken parent;
-				if (customAttrs.Container is IAssemblyDefinition)
+				if (customAttrs.Container is AssemblyDefinition)
 					parent = new MetadataToken (TokenType.Assembly, 1);
-				else if (customAttrs.Container is IModuleDefinition)
+				else if (customAttrs.Container is ModuleDefinition)
 					parent = new MetadataToken (TokenType.Module, 1);
 				else if (customAttrs.Container is IMetadataTokenProvider)
 					parent = (customAttrs.Container as IMetadataTokenProvider).MetadataToken;
@@ -615,13 +615,13 @@ namespace Mono.Cecil {
 		{
 			FieldMarshalTable fmTable = m_tableWriter.GetFieldMarshalTable ();
 			FieldMarshalRow fmRow = m_rowWriter.CreateFieldMarshalRow (
-				((IMetadataTokenProvider) marshalSpec.Container).MetadataToken,
+				marshalSpec.Container.MetadataToken,
 				m_sigWriter.AddMarshalSig (GetMarshalSig (marshalSpec)));
 
 			fmTable.Rows.Add (fmRow);
 		}
 
-		void WriteConstant (IHasConstant hc, MetadataToken parent, TypeReference type)
+		void WriteConstant (IHasConstant hc, TypeReference type)
 		{
 			ConstantTable cTable = m_tableWriter.GetConstantTable ();
 			ElementType et;
@@ -636,7 +636,7 @@ namespace Mono.Cecil {
 
 			ConstantRow cRow = m_rowWriter.CreateConstantRow (
 				et,
-				parent,
+				hc.MetadataToken,
 				m_mdWriter.AddBlob (EncodeConstant (et, hc.Constant)));
 
 			cTable.Rows.Add (cRow);
@@ -646,7 +646,7 @@ namespace Mono.Cecil {
 		{
 			FieldLayoutTable flTable = m_tableWriter.GetFieldLayoutTable ();
 			FieldLayoutRow flRow = m_rowWriter.CreateFieldLayoutRow (
-				field.LayoutInfo.Offset,
+				field.Offset,
 				GetRidFor (field));
 
 			flTable.Rows.Add (flRow);
@@ -656,15 +656,15 @@ namespace Mono.Cecil {
 		{
 			ClassLayoutTable clTable = m_tableWriter.GetClassLayoutTable ();
 			ClassLayoutRow clRow = m_rowWriter.CreateClassLayoutRow (
-				type.LayoutInfo.PackingSize,
-				type.LayoutInfo.ClassSize,
+				type.PackingSize,
+				type.ClassSize,
 				GetRidFor (type));
 
 			clTable.Rows.Add (clRow);
 		}
 
 		void WriteSemantic (MethodSemanticsAttributes attrs,
-			IMemberDefinition member, MethodDefinition meth)
+			IMetadataTokenProvider member, MethodDefinition meth)
 		{
 			MethodSemanticsTable msTable = m_tableWriter.GetMethodSemanticsTable ();
 			MethodSemanticsRow msRow = m_rowWriter.CreateMethodSemanticsRow (
@@ -678,54 +678,106 @@ namespace Mono.Cecil {
 		void SortTables ()
 		{
 			TablesHeap th = m_mdWriter.GetMetadataRoot ().Streams.TablesHeap;
+			th.Sorted = 0;
 
-			if (th.HasTable (typeof (NestedClassTable)))
+			if (th.HasTable (NestedClassTable.RId))
 				m_tableWriter.GetNestedClassTable ().Rows.Sort (
 					TableComparers.NestedClass.Instance);
+			th.Sorted |= ((long) 1 << NestedClassTable.RId);
 
-			if (th.HasTable (typeof (InterfaceImplTable)))
+			if (th.HasTable (InterfaceImplTable.RId))
 				m_tableWriter.GetInterfaceImplTable ().Rows.Sort (
 					TableComparers.InterfaceImpl.Instance);
+			th.Sorted |= ((long) 1 << InterfaceImplTable.RId);
 
-			if (th.HasTable (typeof (ConstantTable)))
+			if (th.HasTable (ConstantTable.RId))
 				m_tableWriter.GetConstantTable ().Rows.Sort (
 					TableComparers.Constant.Instance);
+			th.Sorted |= ((long) 1 << ConstantTable.RId);
 
-			if (th.HasTable (typeof (MethodSemanticsTable)))
+			if (th.HasTable (MethodSemanticsTable.RId))
 				m_tableWriter.GetMethodSemanticsTable ().Rows.Sort (
 					TableComparers.MethodSem.Instance);
+			th.Sorted |= ((long) 1 << MethodSemanticsTable.RId);
 
-			if (th.HasTable (typeof (FieldMarshalTable)))
+			if (th.HasTable (FieldMarshalTable.RId))
 				m_tableWriter.GetFieldMarshalTable ().Rows.Sort (
 					TableComparers.FieldMarshal.Instance);
+			th.Sorted |= ((long) 1 << FieldMarshalTable.RId);
 
-			if (th.HasTable (typeof (ClassLayoutTable)))
+			if (th.HasTable (ClassLayoutTable.RId))
 				m_tableWriter.GetClassLayoutTable ().Rows.Sort (
 					TableComparers.TypeLayout.Instance);
+			th.Sorted |= ((long) 1 << ClassLayoutTable.RId);
 
-			if (th.HasTable (typeof (FieldLayoutTable)))
+			if (th.HasTable (FieldLayoutTable.RId))
 				m_tableWriter.GetFieldLayoutTable ().Rows.Sort (
 					TableComparers.FieldLayout.Instance);
+			th.Sorted |= ((long) 1 << FieldLayoutTable.RId);
 
-			if (th.HasTable (typeof (ImplMapTable)))
+			if (th.HasTable (ImplMapTable.RId))
 				m_tableWriter.GetImplMapTable ().Rows.Sort (
 					TableComparers.PInvoke.Instance);
+			th.Sorted |= ((long) 1 << ImplMapTable.RId);
 
-			if (th.HasTable (typeof (FieldRVATable)))
+			if (th.HasTable (FieldRVATable.RId))
 				m_tableWriter.GetFieldRVATable ().Rows.Sort (
 					TableComparers.FieldRVA.Instance);
+			th.Sorted |= ((long) 1 << FieldRVATable.RId);
 
-			if (th.HasTable (typeof (MethodImplTable)))
+			if (th.HasTable (MethodImplTable.RId))
 				m_tableWriter.GetMethodImplTable ().Rows.Sort (
 					TableComparers.Override.Instance);
+			th.Sorted |= ((long) 1 << MethodImplTable.RId);
 
-			if (th.HasTable (typeof (CustomAttributeTable)))
+			if (th.HasTable (CustomAttributeTable.RId))
 				m_tableWriter.GetCustomAttributeTable ().Rows.Sort (
 					TableComparers.CustomAttribute.Instance);
+			th.Sorted |= ((long) 1 << CustomAttributeTable.RId);
 
-			if (th.HasTable (typeof (DeclSecurityTable)))
+			if (th.HasTable (DeclSecurityTable.RId))
 				m_tableWriter.GetDeclSecurityTable ().Rows.Sort (
 					TableComparers.SecurityDeclaration.Instance);
+			th.Sorted |= ((long) 1 << DeclSecurityTable.RId);
+		}
+
+		void CompleteGenericTables ()
+		{
+			if (m_genericParamStack.Count == 0)
+				return;
+
+			TablesHeap th = m_mdWriter.GetMetadataRoot ().Streams.TablesHeap;
+			GenericParamTable gpTable = m_tableWriter.GetGenericParamTable ();
+			GenericParamConstraintTable gpcTable = m_tableWriter.GetGenericParamConstraintTable ();
+
+			(m_genericParamStack as ArrayList).Sort (TableComparers.GenericParam.Instance);
+
+			foreach (GenericParameter gp in m_genericParamStack) {
+				GenericParamRow gpRow = m_rowWriter.CreateGenericParamRow (
+					(ushort) gp.Owner.GenericParameters.IndexOf (gp),
+					gp.Attributes,
+					gp.Owner.MetadataToken,
+					m_mdWriter.AddString (gp.Name));
+
+				gpTable.Rows.Add (gpRow);
+				gp.MetadataToken = new MetadataToken (TokenType.GenericParam, (uint) gpTable.Rows.Count);
+
+				VisitCustomAttributeCollection (gp.CustomAttributes);
+
+				if (gp.Constraints.Count == 0)
+					continue;
+
+				foreach (TypeReference constraint in gp.Constraints) {
+					GenericParamConstraintRow gpcRow = m_rowWriter.CreateGenericParamConstraintRow (
+						(uint) gpTable.Rows.Count,
+						GetTypeDefOrRefToken (constraint));
+
+					gpcTable.Rows.Add (gpcRow);
+				}
+			}
+
+			th.Sorted |= ((long) 1 << GenericParamTable.RId);
+			th.Sorted |= ((long) 1 << GenericParamConstraintTable.RId);
 		}
 
 		public override void TerminateModuleDefinition (ModuleDefinition module)
@@ -734,18 +786,19 @@ namespace Mono.Cecil {
 			VisitSecurityDeclarationCollection (module.Assembly.SecurityDeclarations);
 			VisitCustomAttributeCollection (module.CustomAttributes);
 
+			CompleteGenericTables ();
 			SortTables ();
 
 			MethodTable mTable = m_tableWriter.GetMethodTable ();
 			for (int i = 0; i < m_methodStack.Count; i++) {
 				MethodDefinition meth = (MethodDefinition) m_methodStack [i];
-				if (!meth.IsAbstract && meth.PInvokeInfo == null)
+				if (meth.HasBody)
 					mTable [i].RVA = m_codeWriter.WriteMethodBody (meth);
 			}
 
 			if (m_fieldStack.Count > 0) {
 				FieldRVATable frTable = null;
-				foreach (IFieldDefinition field in m_fieldStack) {
+				foreach (FieldDefinition field in m_fieldStack) {
 					if (field.InitialValue != null && field.InitialValue.Length > 0) {
 						if (frTable == null)
 							frTable = m_tableWriter.GetFieldRVATable ();
@@ -800,6 +853,8 @@ namespace Mono.Cecil {
 				return ElementType.String;
 			case Constants.Type :
 				return ElementType.Type;
+			case Constants.Object :
+				return ElementType.Object;
 			default:
 				return ElementType.Class;
 			}
@@ -809,42 +864,48 @@ namespace Mono.Cecil {
 		{
 			m_constWriter.Empty ();
 
+			if (value == null)
+				et = ElementType.Class;
+
+			IConvertible ic = value as IConvertible;
+			IFormatProvider fp = CultureInfo.CurrentCulture.NumberFormat;
+
 			switch (et) {
 			case ElementType.Boolean :
-				m_constWriter.Write ((byte) (((bool) value) ? 1 : 0));
+				m_constWriter.Write ((byte) (ic.ToBoolean (fp) ? 1 : 0));
 				break;
 			case ElementType.Char :
-				m_constWriter.Write ((ushort) (char) value);
+				m_constWriter.Write ((ushort) ic.ToChar (fp));
 				break;
 			case ElementType.I1 :
-				m_constWriter.Write ((sbyte) value);
+				m_constWriter.Write (ic.ToSByte (fp));
 				break;
 			case ElementType.I2 :
-				m_constWriter.Write ((short) value);
+				m_constWriter.Write (ic.ToInt16 (fp));
 				break;
 			case ElementType.I4 :
-				m_constWriter.Write ((int) value);
+				m_constWriter.Write (ic.ToInt32 (fp));
 				break;
 			case ElementType.I8 :
-				m_constWriter.Write ((long) value);
+				m_constWriter.Write (ic.ToInt64 (fp));
 				break;
 			case ElementType.U1 :
-				m_constWriter.Write ((byte) value);
+				m_constWriter.Write (ic.ToByte (fp));
 				break;
 			case ElementType.U2 :
-				m_constWriter.Write ((ushort) value);
+				m_constWriter.Write (ic.ToUInt16 (fp));
 				break;
 			case ElementType.U4 :
-				m_constWriter.Write ((uint) value);
+				m_constWriter.Write (ic.ToUInt32 (fp));
 				break;
 			case ElementType.U8 :
-				m_constWriter.Write ((ulong) value);
+				m_constWriter.Write (ic.ToUInt64 (fp));
 				break;
 			case ElementType.R4 :
-				m_constWriter.Write ((float) value);
+				m_constWriter.Write (ic.ToSingle (fp));
 				break;
 			case ElementType.R8 :
-				m_constWriter.Write ((double) value);
+				m_constWriter.Write (ic.ToDouble (fp));
 				break;
 			case ElementType.String :
 				m_constWriter.Write (Encoding.Unicode.GetBytes ((string) value));
@@ -859,7 +920,7 @@ namespace Mono.Cecil {
 			return m_constWriter.ToArray ();
 		}
 
-		public SigType GetSigType (ITypeReference type)
+		public SigType GetSigType (TypeReference type)
 		{
 			string name = type.FullName;
 
@@ -902,7 +963,7 @@ namespace Mono.Cecil {
 				return new SigType (ElementType.TypedByRef);
 			}
 
-			if (type is IGenericParameter) {
+			if (type is GenericParameter) {
 				GenericParameter gp = type as GenericParameter;
 				int pos = gp.Owner.GenericParameters.IndexOf (gp);
 				if (gp.Owner is TypeReference)
@@ -923,8 +984,8 @@ namespace Mono.Cecil {
 					gi.Signature.Types [i] = GetSigType (git.GenericArguments [i]);
 
 				return gi;
-			} else if (type is IArrayType) {
-				IArrayType aryType = type as IArrayType;
+			} else if (type is ArrayType) {
+				ArrayType aryType = type as ArrayType;
 				if (aryType.IsSizedArray) {
 					SZARRAY szary = new SZARRAY ();
 					szary.Type = GetSigType (aryType.ElementType);
@@ -934,30 +995,42 @@ namespace Mono.Cecil {
 				// not optimized
 				ArrayShape shape = new ArrayShape ();
 				shape.Rank = aryType.Dimensions.Count;
-				shape.Sizes = new int [shape.Rank];
-				shape.LoBounds = new int [shape.Rank];
+				shape.NumSizes = 0;
+
+				for (int i = 0; i < shape.Rank; i++) {
+					ArrayDimension dim = aryType.Dimensions [i];
+					if (dim.UpperBound > 0)
+						shape.NumSizes++;
+				}
+
+				shape.Sizes = new int [shape.NumSizes];
+				shape.NumLoBounds = shape.Rank;
+				shape.LoBounds = new int [shape.NumLoBounds];
 
 				for (int i = 0; i < shape.Rank; i++) {
 					ArrayDimension dim = aryType.Dimensions [i];
 					shape.LoBounds [i] = dim.LowerBound;
-					shape.Sizes [i] = dim.UpperBound - dim.LowerBound + 1;
+					if (dim.UpperBound > 0)
+						shape.Sizes [i] = dim.UpperBound - dim.LowerBound + 1;
 				}
 
 				ARRAY ary = new ARRAY ();
 				ary.Shape = shape;
 				ary.Type = GetSigType (aryType.ElementType);
 				return ary;
-			} else if (type is IPointerType) {
+			} else if (type is PointerType) {
 				PTR p = new PTR ();
-				ITypeReference elementType = (type as IPointerType).ElementType;
+				TypeReference elementType = (type as PointerType).ElementType;
 				p.Void = elementType.FullName == Constants.Void;
 				if (!p.Void) {
 					p.CustomMods = GetCustomMods (elementType);
 					p.PtrType = GetSigType (elementType);
 				}
 				return p;
-			} else if (type is IFunctionPointerType) {
+			} else if (type is FunctionPointerType) {
 				throw new NotImplementedException ("Function pointer are not implemented"); // TODO
+			} else if (type is TypeSpecification) {
+				return GetSigType ((type as TypeSpecification).ElementType);
 			} else if (type.IsValueType) {
 				VALUETYPE vt = new VALUETYPE ();
 				vt.Type = GetTypeDefOrRefToken (type);
@@ -969,41 +1042,41 @@ namespace Mono.Cecil {
 			}
 		}
 
-		public CustomMod [] GetCustomMods (ITypeReference type)
+		public CustomMod [] GetCustomMods (TypeReference type)
 		{
-			if (!(type is IModifierType))
+			if (!(type is ModType))
 				return new CustomMod [0];
 
 			ArrayList cmods = new ArrayList ();
-			ITypeReference cur = type;
-			while (cur is IModifierType) {
-				if (cur is IModifierOptional) {
+			TypeReference cur = type;
+			while (cur is ModType) {
+				if (cur is ModifierOptional) {
 					CustomMod cmod = new CustomMod ();
 					cmod.CMOD = CustomMod.CMODType.OPT;
-					cmod.TypeDefOrRef = GetTypeDefOrRefToken ((cur as IModifierOptional).ModifierType);
+					cmod.TypeDefOrRef = GetTypeDefOrRefToken ((cur as ModifierOptional).ModifierType);
 					cmods.Add (cmod);
-				} else if (cur is IModifierRequired) {
+				} else if (cur is ModifierRequired) {
 					CustomMod cmod = new CustomMod ();
 					cmod.CMOD = CustomMod.CMODType.REQD;
-					cmod.TypeDefOrRef = GetTypeDefOrRefToken ((cur as IModifierRequired).ModifierType);
+					cmod.TypeDefOrRef = GetTypeDefOrRefToken ((cur as ModifierRequired).ModifierType);
 					cmods.Add (cmod);
 				}
 
-				cur = (cur as IModifierType).ElementType;
+				cur = (cur as ModType).ElementType;
 			}
 
 			return cmods.ToArray (typeof (CustomMod)) as CustomMod [];
 		}
 
-		public Signature GetMemberRefSig (IMemberReference member)
+		public Signature GetMemberRefSig (MemberReference member)
 		{
-			if (member is IFieldReference)
-				return GetFieldSig (member as IFieldReference);
+			if (member is FieldReference)
+				return GetFieldSig (member as FieldReference);
 			else
-				return GetMemberRefSig (member as IMethodReference);
+				return GetMemberRefSig (member as MethodReference);
 		}
 
-		public FieldSig GetFieldSig (IFieldReference field)
+		public FieldSig GetFieldSig (FieldReference field)
 		{
 			FieldSig sig = new FieldSig ();
 			sig.CallingConvention |= 0x6;
@@ -1013,19 +1086,19 @@ namespace Mono.Cecil {
 			return sig;
 		}
 
-		Param [] GetParametersSig (IParameterDefinitionCollection parameters)
+		Param [] GetParametersSig (ParameterDefinitionCollection parameters)
 		{
 			Param [] ret = new Param [parameters.Count];
 			for (int i = 0; i < ret.Length; i++) {
-				IParameterDefinition pDef =parameters [i];
+				ParameterDefinition pDef = parameters [i];
 				Param p = new Param ();
 				p.CustomMods = GetCustomMods (pDef.ParameterType);
 				if (pDef.ParameterType.FullName == Constants.TypedReference)
 					p.TypedByRef = true;
-				else if (pDef.ParameterType is IReferenceType) {
+				else if (pDef.ParameterType is ReferenceType) {
 					p.ByRef = true;
 					p.Type = GetSigType (
-						(pDef.ParameterType as IReferenceType).ElementType);
+						(pDef.ParameterType as ReferenceType).ElementType);
 				} else
 					p.Type = GetSigType (pDef.ParameterType);
 				ret [i] = p;
@@ -1033,7 +1106,7 @@ namespace Mono.Cecil {
 			return ret;
 		}
 
-		void CompleteMethodSig (IMethodReference meth, MethodSig sig)
+		void CompleteMethodSig (MethodReference meth, MethodSig sig)
 		{
 			sig.HasThis = meth.HasThis;
 			sig.ExplicitThis = meth.ExplicitThis;
@@ -1055,17 +1128,17 @@ namespace Mono.Cecil {
 				rtSig.Void = true;
 			else if (meth.ReturnType.ReturnType.FullName == Constants.TypedReference)
 				rtSig.TypedByRef = true;
-			else if (meth.ReturnType.ReturnType is IReferenceType) {
+			else if (meth.ReturnType.ReturnType is ReferenceType) {
 				rtSig.ByRef = true;
 				rtSig.Type = GetSigType (
-					(meth.ReturnType.ReturnType as IReferenceType).ElementType);
+					(meth.ReturnType.ReturnType as ReferenceType).ElementType);
 			} else
 				rtSig.Type = GetSigType (meth.ReturnType.ReturnType);
 
 			sig.RetType = rtSig;
 		}
 
-		public MethodRefSig GetMethodRefSig (IMethodReference meth)
+		public MethodRefSig GetMethodRefSig (MethodReference meth)
 		{
 			if (meth.GenericParameters.Count > 0)
 				return GetMethodDefSig (meth);
@@ -1086,7 +1159,7 @@ namespace Mono.Cecil {
 			return methSig;
 		}
 
-		public MethodDefSig GetMethodDefSig (IMethodReference meth)
+		public MethodDefSig GetMethodDefSig (MethodReference meth)
 		{
 			MethodDefSig sig = new MethodDefSig ();
 
@@ -1100,7 +1173,7 @@ namespace Mono.Cecil {
 			return sig;
 		}
 
-		public PropertySig GetPropertySig (IPropertyDefinition prop)
+		public PropertySig GetPropertySig (PropertyDefinition prop)
 		{
 			PropertySig ps = new PropertySig ();
 			ps.CallingConvention |= 0x8;
@@ -1108,9 +1181,9 @@ namespace Mono.Cecil {
 			bool hasThis;
 			bool explicitThis;
 			MethodCallingConvention mcc;
-			IParameterDefinitionCollection parameters = prop.Parameters;
+			ParameterDefinitionCollection parameters = prop.Parameters;
 
-			IMethodDefinition meth;
+			MethodDefinition meth;
 			if (prop.GetMethod != null)
 				meth = prop.GetMethod;
 			else if (prop.SetMethod != null)
@@ -1145,7 +1218,7 @@ namespace Mono.Cecil {
 			return ps;
 		}
 
-		public TypeSpec GetTypeSpecSig (ITypeReference type)
+		public TypeSpec GetTypeSpecSig (TypeReference type)
 		{
 			TypeSpec ts = new TypeSpec (GetSigType (type));
 			return ts;
@@ -1162,48 +1235,120 @@ namespace Mono.Cecil {
 			return new MethodSpec (gis);
 		}
 
-		public CustomAttrib GetCustomAttributeSig (ICustomAttribute ca)
+		string GetObjectTypeName (object o)
+		{
+			Type t = o.GetType ();
+			return string.Concat (t.Namespace, ".", t.Name);
+		}
+
+		CustomAttrib.Elem CreateElem (TypeReference type, object value)
+		{
+			CustomAttrib.Elem elem = new CustomAttrib.Elem ();
+			elem.Value = value;
+			elem.ElemType = type;
+			elem.FieldOrPropType = GetCorrespondingType (type.FullName);
+
+			switch (elem.FieldOrPropType) {
+			case ElementType.Boolean :
+			case ElementType.Char :
+			case ElementType.R4 :
+			case ElementType.R8 :
+			case ElementType.I1 :
+			case ElementType.I2 :
+			case ElementType.I4 :
+			case ElementType.I8 :
+			case ElementType.U1 :
+			case ElementType.U2 :
+			case ElementType.U4 :
+			case ElementType.U8 :
+				elem.Simple = true;
+				break;
+			case ElementType.String:
+				elem.String = true;
+				break;
+			case ElementType.Type:
+				elem.Type = true;
+				break;
+			case ElementType.Object:
+				elem.BoxedValueType = true;
+				if (value == null)
+					elem.FieldOrPropType = ElementType.String;
+				else
+					elem.FieldOrPropType = GetCorrespondingType (
+						GetObjectTypeName (value));
+				break;
+			}
+
+			return elem;
+		}
+
+		CustomAttrib.FixedArg CreateFixedArg (TypeReference type, object value)
+		{
+			CustomAttrib.FixedArg fa = new CustomAttrib.FixedArg ();
+			if (value is object []) {
+				fa.SzArray = true;
+				object [] values = value as object [];
+				TypeReference obj = ((ArrayType) type).ElementType;
+				fa.NumElem = (uint) values.Length;
+				fa.Elems = new CustomAttrib.Elem [values.Length];
+				for (int i = 0; i < values.Length; i++)
+					fa.Elems [i] = CreateElem (obj, values [i]);
+			} else {
+				fa.Elems = new CustomAttrib.Elem [1];
+				fa.Elems [0] = CreateElem (type, value);
+			}
+
+			return fa;
+		}
+
+		CustomAttrib.NamedArg CreateNamedArg (TypeReference type, string name,
+			object value, bool field)
+		{
+			CustomAttrib.NamedArg na = new CustomAttrib.NamedArg ();
+			na.Field = field;
+			na.Property = !field;
+
+			na.FieldOrPropName = name;
+			na.FieldOrPropType = GetCorrespondingType (type.FullName);
+			na.FixedArg = CreateFixedArg (type, value);
+
+			return na;
+		}
+
+		public CustomAttrib GetCustomAttributeSig (CustomAttribute ca)
 		{
 			CustomAttrib cas = new CustomAttrib (ca.Constructor);
 			cas.Prolog = CustomAttrib.StdProlog;
 
-			cas.FixedArgs = new CustomAttrib.FixedArg [0];
-
 			cas.FixedArgs = new CustomAttrib.FixedArg [ca.Constructor.Parameters.Count];
 
-			for (int i = 0; i < cas.FixedArgs.Length; i++) {
-				object o = ca.ConstructorParameters [i];
-				CustomAttrib.FixedArg fa = new CustomAttrib.FixedArg ();
-//				if (o is object []) {
-//					object [] values = o as object [];
-//					fa.Elems = new CustomAttrib.Elem [values.Length];
-//					for (int j = 0; j < values.Length; j++) {
-//						CustomAttrib.Elem elem = new CustomAttrib.Elem ();
-//						elem.Value = values [j];
-//						elem.FieldOrPropType = ElementType.Object;
-//						elem.ElemType = ca.Constructor.Parameters [i].ParameterType;
-//						fa.Elems [j] = elem;
-//					}
-//				} else {
-					fa.Elems = new CustomAttrib.Elem [1];
-					fa.Elems [0].Value = o;
-					fa.Elems [0].ElemType = ca.Constructor.Parameters [i].ParameterType;
-					fa.Elems [0].FieldOrPropType = GetCorrespondingType (fa.Elems [0].ElemType.FullName);
-					if (fa.Elems [0].FieldOrPropType == ElementType.Class)
-						fa.Elems [0].FieldOrPropType = ElementType.I4; // buggy
-//				}
+			for (int i = 0; i < cas.FixedArgs.Length; i++)
+				cas.FixedArgs [i] = CreateFixedArg (
+					ca.Constructor.Parameters [i].ParameterType, ca.ConstructorParameters [i]);
 
-				cas.FixedArgs [i] = fa;
+			int nn = ca.Fields.Count + ca.Properties.Count;
+			cas.NumNamed = (ushort) nn;
+			cas.NamedArgs = new CustomAttrib.NamedArg [nn];
+
+			if (cas.NamedArgs.Length > 0) {
+				int curs = 0;
+				foreach (DictionaryEntry entry in ca.Fields) {
+					string field = (string) entry.Key;
+					cas.NamedArgs [curs++] = CreateNamedArg (
+						ca.GetFieldType (field), field, entry.Value, true);
+				}
+
+				foreach (DictionaryEntry entry in ca.Properties) {
+					string property = (string) entry.Key;
+					cas.NamedArgs [curs++] = CreateNamedArg (
+						ca.GetPropertyType (property), property, entry.Value, false);
+				}
 			}
 
-			cas.NumNamed = 0;
-
-			cas.NamedArgs = new CustomAttrib.NamedArg [0];
-
 			return cas;
 		}
 
-		public MarshalSig GetMarshalSig (IMarshalSpec mSpec)
+		public MarshalSig GetMarshalSig (MarshalDesc mSpec)
 		{
 			MarshalSig ms = new MarshalSig (mSpec.NativeIntrinsic);
 
@@ -1220,7 +1365,7 @@ namespace Mono.Cecil {
 				MarshalSig.CustomMarshaler cm = new MarshalSig.CustomMarshaler ();
 				cm.Guid = cmd.Guid.ToString ();
 				cm.UnmanagedType = cmd.UnmanagedType;
-				cm.ManagedType = cmd.ManagedType.FullName;
+				cm.ManagedType = cmd.ManagedType;
 				cm.Cookie = cmd.Cookie;
 				ms.Spec = cm;
 			} else if (mSpec is FixedArrayDesc) {
